package meerkat.mixer.mixing;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.Encryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.crypto.mixnet.MixerOutput;
import meerkat.protobuf.Crypto.EncryptionRandomness;
import meerkat.protobuf.Crypto.RerandomizableEncryptedMessage;
import meerkat.protobuf.Mixing.ZeroKnowledgeProof;

import java.util.List;
import java.util.Random;

/**
 * an implementation of meerkat.crypto.mixnet.Mixer
 * meerkat.mixer.mixing algorithm on set of n encrypted votes:
 *  0. asset n is power of two
 *  1. set switches according to benes network on random permutation
 *  2. re encrypt and mix with respect to switches values (encryptor.rerandomize)
 *  3. generate zero knowledge proof on each re encrypted couple
 *  4. return proofs table + encryption table
 *
 */
public class Mixer implements meerkat.crypto.mixnet.Mixer {

    private final Mix2ZeroKnowledgeProver prover;
    private final Encryption encryptor;

    /**
     *  constructor
     * @param prover
     * @param encryptor
     */
    public Mixer(Mix2ZeroKnowledgeProver prover, Encryption encryptor) {
        this.prover = prover;
        this.encryptor = encryptor;
    }

    /**
     * return True iff n == 2 ^ k
     * @param n
     * @return
     */
    public static boolean isPowerOfTwo(int n){
        return (n & (n - 1)) == 0;
    }

    /**
     * initialize encryption table of size (layers + 1)* n
     * @param n number of votes
     * @param layers
     * @param ciphertexts encrypted votes
     * @return an initialized encryption table s.t first layer == given encrypted votes
     */
    private RerandomizableEncryptedMessage[][] initializeEncryptionTable(int n,int layers,List<RerandomizableEncryptedMessage> ciphertexts){
        // set first level of encryption
        RerandomizableEncryptedMessage[][] encryptionTable = new RerandomizableEncryptedMessage[layers + 1][n];
        for (int j = 0; j < n; j++) {
            encryptionTable[0][j] = ciphertexts.get(j);
        }
        return  encryptionTable;
    }

    /**
     * generate randomness for all rerandomize operations
     * @param n number of votes
     * @param layers
     * @return an initialized randomness table of size layers * n
     */
    private  EncryptionRandomness[][] generateRandomnessesForRerandomize(int n,int layers,Random random){
        EncryptionRandomness[][] randomnesses = new EncryptionRandomness[layers][n];
        for (int layer = 0; layer < layers; layer++)
        {
            for (int i = 0; i < n; i++) {
                randomnesses[layer][i] = encryptor.generateRandomness(random);;
            }
        }
        return randomnesses;
    }

    /**
     * generate new random mix network
     * @param n number of votes
     * @param random
     * @return new random mix network
     */
    private MixNetwork generateMixNetwork(int n,Random random){
        return new MixNetwork(new RandomPermutation(n,random));
    }

    /**
     * fills the encryption table with rerandomize encrypted votes.
     * @param layers
     * @param mixNetwork switches table (boolean values)
     * @param encryptionTable an initialized encryption table s.t first layer == given encrypted votes
     * @param randomnesses an initialized randomness table of size layers * n, use for rerandomize operations
     * @throws InvalidProtocolBufferException
     */
    private void rerandomize(int layers,MixNetwork mixNetwork, RerandomizableEncryptedMessage[][] encryptionTable
            ,EncryptionRandomness[][] randomnesses) throws InvalidProtocolBufferException {
        Switch[] switchesLayer;
        int index1,index2;
        RerandomizableEncryptedMessage e1,e2;
        EncryptionRandomness r1,r2;
        for (int layer = 0; layer < layers; layer++)
        {
            switchesLayer = mixNetwork.getSwitchesByLayer(layer);
            for (Switch sw : switchesLayer) {
                index1 = sw.i;
                index2 = sw.j;
                e1 = encryptionTable[layer][index1];
                e2 = encryptionTable[layer][index2];

                r1 = randomnesses[layer][index1];
                r2 = randomnesses[layer][index2];
                if (!sw.value) {
                    encryptionTable[layer + 1][index1] = encryptor.rerandomize(e1, r1);
                    encryptionTable[layer + 1][index2] = encryptor.rerandomize(e2, r2);

                } else {
                    encryptionTable[layer + 1][index1] = encryptor.rerandomize(e2, r2);
                    encryptionTable[layer + 1][index2] = encryptor.rerandomize(e1, r1);
                }
            }
        }
    }

    /**
     *  generate zero knowledge proof for each rerandomize encrypted votes couple in encryptionTable
     * @param n number of votes
     * @param layers
     * @param mixNetwork switches table (boolean values) used for set encryption table
     * @param encryptionTable full encryption table
     * @param randomnesses randomness table of size layers * n, used for set encryption table
     * @return zero knowledge proofs table
     * @throws InvalidProtocolBufferException
     */
    private ZeroKnowledgeProof[][] generateZeroKnowledgeProofTable(int n, int layers, MixNetwork mixNetwork
            , RerandomizableEncryptedMessage[][] encryptionTable
            , EncryptionRandomness[][] randomnesses) throws InvalidProtocolBufferException {
        Switch[] switchesLayer;
        int index1,index2;
        int switchIndex = 0;
        int nDiv2 = n >> 1;
        ZeroKnowledgeProof[][] proofsTable = new ZeroKnowledgeProof[layers][nDiv2];

        RerandomizableEncryptedMessage e1,e2;
        EncryptionRandomness r1,r2;
        for (int layer = 0; layer < layers; layer++)
        {
            switchesLayer = mixNetwork.getSwitchesByLayer(layer);
            for (Switch sw : switchesLayer) {
                index1 = sw.i;
                index2 = sw.j;
                e1 = encryptionTable[layer][index1];
                e2 = encryptionTable[layer][index2];
                r1 = randomnesses[layer][index1];
                r2 = randomnesses[layer][index2];

                proofsTable[layer][switchIndex] =
                        prover.prove(e1, e2, encryptionTable[layer + 1][index1],
                                encryptionTable[layer + 1][index2],
                                sw.value, sw.i, sw.j, sw.layer, r1, r2);

                switchIndex = (switchIndex + 1) % nDiv2;
            }
        }
        return proofsTable;
    }

    /**
     * mix given encrypted votes using random
     * @param ciphertexts encrypted votes
     * @param random
     * @return meerkat.mixer.mixing result
     * @throws InvalidProtocolBufferException
     */
    public MixerOutput mix(List<RerandomizableEncryptedMessage> ciphertexts,Random random) throws InvalidProtocolBufferException {

        int n = ciphertexts.size();
        assert (n > 1 && isPowerOfTwo(n));

        int layers = MixNetwork.numberOfLayers(n);  // layers = 2logn -1
        RerandomizableEncryptedMessage[][] encryptionTable = initializeEncryptionTable(n,layers,ciphertexts);
        EncryptionRandomness[][] randomnesses = generateRandomnessesForRerandomize(n,layers,random);
        MixNetwork mixNetwork = generateMixNetwork(n,random);
        rerandomize(layers,mixNetwork,encryptionTable,randomnesses);
        ZeroKnowledgeProof[][] proofsTable = generateZeroKnowledgeProofTable(n,layers,mixNetwork,encryptionTable,randomnesses);

        return new meerkat.mixer.mixing.MixerOutput(n,layers,proofsTable, encryptionTable);
    }

}